/**
**  @page WindowUnixIo
**
**  The header file, WindowsUnixIO defines symbols that simualte the level 1 i/o features of unix
**  but:
**    @li Windows File I/O calls are used to implement them
**    @li They are ensconsed in a C++ class.
**
**  They are only available on windows -- on unix just use the unix versions -- or if you need portability, duplicate this
**  class 
**/
#ifndef CXXTLS_WINDOWS_UNIX_IO_INCLUDED_P
#define CXXTLS_WINDOWS_UNIX_IO_INCLUDED_P
#ifdef _MSC_VER

#include <string>

#include <Windows.h>

class WindowsFileHandle
    /// a wrapper around MS Windows HANDLEs that give you unix style I/O operations
    /// conveniently packaged.
{
      HANDLE fd_; 
      bool   error_;

      
public:

      bool error() const { return error_; }


      struct FileStatus
      {
          long long size_;     // list should grow over time as needed
                               // I'm currently sure how to convert windows file times to unix times...
      };


      WindowsFileHandle()
      :  fd_(INVALID_HANDLE_VALUE)
      ,  error_(true)
      {
      }

      ~WindowsFileHandle()
      {
          close(); // safe to close even if not open
      }

      enum BIT_NAMES
           ///  The names of bits passed to the various open options.  Combine these together in BIT_FIELDS using operator |
      {
           //
           // access options
           //
           OM_READ=1,                  ///< open for read
           OM_WRITE=2,                 ///< open for write

           //
           // sharing options
           //
           OM_OTHERS_CAN_READ=4,       ///< allow others to read 
           OM_OTHERS_CAN_WRITE=8,      ///< allow others to write
           OM_OTHERS_CAN_DELETE=16,    ///< allow others to delete 

           //
           //  creation options
           //
           OM_OPEN_EXISTING_ONLY,      ///< only open pre-existing files.  Superseeds OM_OPEN_ANY, and OPEN_NEW_ONLY
           OM_OPEN_NEW_ONLY,           ///< fail if the file exists already.  Superseeds OM_OPEN_ANY
           OM_OPEN_ANY,                ///< open or create a new file.  See OM_TRUNCATE
           OM_TRUNCATE,                ///< if opening an existing file, truncate it.
      };

      typedef unsigned int BIT_FIELD;

      void adopt(HANDLE fd)
          /// take over control of a windows file handle.  This only works if the intended use of the handle,
          /// fd, matches the open mode when fd was created.  See adoptStdin(), etc.
      {
          close();

          error_ = false;

          fd_ = fd;
      }

      void adoptStdin()
         ///  perform unix style i/o operations on the current windows stdin file handle
      {
         adopt(  GetStdHandle(STD_INPUT_HANDLE) );
      }

      void adoptStdout()
         ///  perform unix style i/o operations on the current windows stdout file handle
      {
         adopt(  GetStdHandle(STD_OUTPUT_HANDLE) );
      }

      void adoptStderr()
         ///  perform unix style i/o operations on the current windows stderr file handle
      {
         adopt(  GetStdHandle(STD_ERROR_HANDLE) );
      }




      bool open(std::string const &filename, 
                BIT_FIELD          openMode=OM_READ, 
                BIT_FIELD          shareMode=OM_OTHERS_CAN_READ | OM_OTHERS_CAN_WRITE | OM_OTHERS_CAN_DELETE,
                BIT_FIELD          createMode=OM_OPEN_ANY | OM_TRUNCATE
               )
          /// open the file with a given filename.  Binary input can be read or written to the file
          ///
          /// @return true if successfull. false otherwise.
          ///
          /// @param[in]  filename The name of the file
          /// @param[in]  openMode A bit field containing one or more of OM_READ, OM_WRITE.  The default is OM_READ | OM_WRITE.
          /// @param[in]  shareMode A bit field defining what you want to allow other processes to do. The default is to allow everything.
          /// @param[in]  createMode A bit field defining whether pre-existing files get opened or not and whether their contents get truncated.
      {
          close(); // safe to close even if not open

          DWORD dwDesiredAccess = 0;
          DWORD dwShareMode = 0;
          DWORD dwCreateDisposition = 0;
          DWORD dwFileAttributes = FILE_ATTRIBUTE_NORMAL;

          if(openMode & OM_READ)
             dwDesiredAccess |= GENERIC_READ;

          if(openMode & OM_WRITE)
             dwDesiredAccess |= GENERIC_WRITE;
          
          if(shareMode & OM_OTHERS_CAN_READ)
              dwShareMode |= FILE_SHARE_READ;     

          if(shareMode & OM_OTHERS_CAN_WRITE)     
              dwShareMode |= FILE_SHARE_WRITE;

          if(shareMode & OM_OTHERS_CAN_DELETE)
              dwShareMode |= FILE_SHARE_DELETE;  

          if( (openMode & OM_WRITE) && (createMode &  (OM_OPEN_ANY | OM_OPEN_EXISTING_ONLY)) && (createMode & OM_TRUNCATE))
          {
             // we are trying to write to the file
             // and we are possibly opening an existing file


             dwCreateDisposition |= TRUNCATE_EXISTING;
          }

          if( openMode & OM_OPEN_EXISTING_ONLY)
          {
              dwCreateDisposition |= OPEN_EXISTING;
          }
          else
          {
             // if we get here, we are not opening ONLY existing files

             if(openMode & OM_OPEN_NEW_ONLY)
             {
                dwCreateDisposition |= CREATE_NEW;   // new files only
             }
             else
             {
                // ASSUME OM_OPEN_ANY.  If you don't have it set, we are going to assume that you do.
                dwCreateDisposition |= CREATE_ALWAYS;
             }
          }
              
          fd_ = CreateFile(filename.c_str(), dwDesiredAccess, dwShareMode, 0, dwCreateDisposition, dwFileAttributes, 0);

          error_ = (fd_ == INVALID_HANDLE_VALUE);

         return !error_;  

      }

      void close()
         /// close the file.  it is safe to close mutliple times.
      {
         if(fd_ != INVALID_HANDLE_VALUE)
            CloseHandle(fd_);

         fd_ = INVALID_HANDLE_VALUE;

         error_ = true;

      }


      int read(char *buffer, int length)
          /// read bytes from the file, if it is open and not in an error state
          ///
          /// @return A positive number if at least partially successful.  The return value is the
          ///         the number of bytes read.  If return value is > 0, success even if value is not equal
          ///         to length.
          ///
          /// @param[out] buffer  The location to which data will be stored.
          /// @param[in]  length  The maximum size of the buffer
      {
          if(error_)
             return -1;

          if(fd_ == INVALID_HANDLE_VALUE)
          {
             error_ = true;
             return -1;
          }

          DWORD bytesRead;
          
          if(!ReadFile(fd_, buffer, length, &bytesRead, 0))
          {
             // read error occurred.

             close();

             return -1;

          }

          return bytesRead;


      }

      int write(char *buffer, int length)
          /// write to the file.  Note that the return value may be less than length without this
          /// signifying an error -- but it usually is an error.  You could try to write the remaining
          /// bytes later, but who knows whether it would work or not.
          ///
          /// @return the number of bytes written.  if the number is less than 0 an error occurred.
          ///
          ///
      {
          DWORD bytesWritten= 0;

          if(!WriteFile(fd_, buffer, length, &bytesWritten, 0))
          {
             close();

             return -1;
          }

          return bytesWritten;
      }

      bool getStatus(FileStatus &status)
          /// get the file status using its file handle.
          ///
          /// @return true if successful.  See output variable status for the desired info.
      {
         if(error_)
            return false;


         BY_HANDLE_FILE_INFORMATION info;

         if(!GetFileInformationByHandle(fd_, &info))
         {
           return false;
         }

         status.size_ = info.nFileSizeHigh * ((long long)0x100000000) + info.nFileSizeLow;

         return true;

          
      }


      bool seek(long long targetDestination)
         /// move the file pointer to the specific location.
         /// @return true if successfull.
      {
          if(error_)
              return false;


          DWORD lowWord =  targetDestination & 0xffffffff;
          LONG  highWord= (targetDestination >>32) & 0xffffffff;

          DWORD dwPtrLow = SetFilePointer(  fd_,
                                            lowWord, 
                                            &highWord, 
                                            FILE_BEGIN ); 
            
          // Test for failure
          if ( dwPtrLow == INVALID_SET_FILE_POINTER && 
               GetLastError() != NO_ERROR )
           {
            // Deal with failure
            // . . .

            close();

            return false;
        
           } // End of error handler



          return true;
      }

};


///
///  tailFile is a function that performs a unix "tail" command style of display
///  for a given filename.  You basically given it the name of a file to perform
///  a tail operation on, the number of lines to display, and an output file stream
///  on which to perform the tail operation.
///
///  @param sout -- the wrapper for the output file handle
///  @param follow -- the flag that says to repeatedly display new data added to the file after the initial
///                   lines are displayed.  the count of lines is in @ref linesToDisplayAtFirst
///  @param linesToDisplayAtFirst -- the count of lines to display in the inital burst
///  @param filename -- the name of the file to "follow".
///
///  @return nothing -- mostly because the function won't return unless you control c out of the application
///
///  Note if the application intercepts control c then it should superseed the function cxxtls::shuttingDown()
///  whose default implementation is found in lib/thisapplication.cpp.  Its declared in cxxtls/ThisApplication.h.
///
///  Basically, the function should only return true if the ^C has been pressed.  The function can start returning
///  false again as soon as it likes, but you should wait for tailFile() to finish before changing your mind...
///

extern void tailFile(WindowsFileHandle &sout, 
                     bool               follow,  
                     size_t             linesToDisplayAtFirst, 
                     std::string const &filename
                    );  


#endif
#endif
