#ifndef cxxtls_geometry_point_h_included_p
#define cxxtls_geometry_point_h_included_p

///@file
/// Classes and algorithms for geometric operations.  See
///  @li cxxtls::geometry::Point
///
/// Note that the definitions here are double precision floats, not integers.  This stuff is meant for
/// mathematical algorithms, not so much for cursor windowing activities.
///

#include <cxxtls/foreach.h>
#include <math.h>
#include <stdlib.h>

namespace cxxtls
{
namespace geometry
    ///  The subset of the cxxtls (CXX toolkit) that deals with geometric
    ///  data structures and algorithms.
{


    template<bool DimensionCountMatches>  struct GuaranteePointDimensionCount;
        ///<  A data structure which is constructed only for the purpose of
        ///   producing a compile error if the template parameter is not "true".
        ///  presumably, the points of construction will include an expression
        ///  as the template parameter to evaluable the number of dimensions actually
        /// used in a point matches expectations encoded in the code.

    template<> struct GuaranteePointDimensionCount<true>
        /// Accept proper dimension count.  Note there is no instantiation for
        /// FALSE!  This is done on purpose.
    {
       GuaranteePointDimensionCount(){}
    };

    //------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------
    
    template<int Dimensions>
    class Point
        /// A representation of a point in n-dimensional space.
        ///
        /// The dimensions are not named (ie, you won't find get_x()), but
        /// rather they are numbered (so get_x() = get(0)).
        ///
        /// A point can be considered to be vector from 0,0 to the point.  To get
        /// the magnitude of this vector, call magnitude().
        ///
        /// Dimensional values are double precision float, not int.  This kind of point is
        /// meant for mathematical operations, not cursor windowing.
    {
        double dims_[Dimensions];

    public:

        enum constants_ { 
                          dimensions = Dimensions /// the count of dimensions in this point in n-space
                        };


        double const &get(size_t dimension) const { return dims_[dimension]; }
            ///<  get a dimensional value from this point.  The dimensions are
            ///   are array elements, so x = 0, y = 1, z = 2, ...

        void set(size_t dimension, double value)  { dims_[dimension] = value; }
            ///<  set a dimensional value from this point.  The dimensions are
            ///   are array elements, so x = 0, y = 1, z = 2, ...
            ///   Note that this function sets part of the location in n-dimensional space
            ///   not the value of the data point at that location.

        Point()
            /// construct an uninitialized point.  Note that the dimensions are all
            /// trash.  See member zero() to force all the dimensions of this point
            /// to zero.
        {
        }

        void zero() 
            /// set all the dimensional values to 0.
        {
           CXXTLS_FOREACH(double &value, dims_)
           {
              value = 0;
           }
        }

        Point( double const *input )
            ///  Construct a point by copying an array of dimensions values whose length is
            ///  assumed to be Dimensions in size.
        {
           CXXTLS_FOREACH(double &value, dims_)
           {
              value = *input++;
           }

        }

        Point( double x )
           /// construct a one dimensional point.  
        {
           GuaranteePointDimensionCount< Dimensions == 1> guarantee;
               // if the above line fails to compile, your Dimension count is wrong!
           
           dims_[0] = x;
        }

        Point( double x, double y )
           /// construct a two dimensional point
        {
           GuaranteePointDimensionCount< Dimensions == 2> guarantee; 
               // if the above line fails to compile, your Dimension count is wrong!
                                                                     

           dims_[0] = x;
           dims_[1] = y;
        }

        Point( double x, double y, double z )
           /// construct a three dimensional point
        {
           GuaranteePointDimensionCount< Dimensions == 3> guarantee;
               // if the above line fails to compile, your Dimension count is wrong!

           dims_[0] = x;
           dims_[1] = y;
           dims_[2] = z;
        }

        Point( double x, double y, double z, double a )
           /// construct a four dimensional point
        {
           GuaranteePointDimensionCount< Dimensions == 4> guarantee;
               // if the above line fails to compile, your Dimension count is wrong!

           dims_[0] = x;
           dims_[1] = y;
           dims_[2] = z;
           dims_[3] = a;
        }

        Point( double x, double y, double z, double a, double b )
           /// construct a five dimensional point
        {
           GuaranteePointDimensionCount< Dimensions == 5> guarantee;
               // if the above line fails to compile, your Dimension count is wrong!

           dims_[0] = x;
           dims_[1] = y;
           dims_[2] = z;
           dims_[3] = a;
        }

        template<class Ostream>
        void output(Ostream &s) const
            ///  write this point to a stream
        {
            s  << Dimensions << "[";
    
            int count=0;
    
            CXXTLS_FOREACH(double v, dims_)
            {
                s << v;
    
                ++count;
    
                if(count != Dimensions)
                   s << ",";
               
            }
    
            s << "]";
    
        }

        double distanceTo( Point<Dimensions> const &other) const
            /// @return the distance between this point and the other
        {
            double const *input = &other.dims_[0];
        
            double sum = 0;

            CXXTLS_FOREACH(double v, dims_)
            {
                double delta = v - *input++;

                sum += delta * delta;

            }

            return sqrt(sum);
        }

        double magnitude() const
            /// @return the distance between this point and the origin of the coordinate 
            ///         system
        {
            double sum = 0;

            CXXTLS_FOREACH(double v, dims_)
            {
                sum += v * v;

            }

            return sqrt(sum);
        }




        Point<Dimensions> lesser(Point<Dimensions> const &comparand) const
            /// @return a point which is the lesser of each the dimensions in *this 
            /// as compared to the comparand point.
            ///
            /// @note that this function operates on a per dimension basis rather than
            /// gestalt basis.
        {
           Point<Dimensions> rv = *this;

           for(size_t i = 0; i < Dimensions; ++ i)
           {
              if( comparand.get(i) < rv.get(i))
              {
                 rv.set(i, comparand.get(i));
              }
           }

           return rv;

        }

        Point<Dimensions> greater(Point<Dimensions> const &comparand) const
            /// @return a point which is the greater of each the dimensions in *this 
            /// as compared to the comparand point.
            ///
            /// @note that this function operates on a per dimension basis rather than
            /// gestalt basis.
        {
           Point<Dimensions> rv = *this;

           for(size_t i = 0; i < Dimensions; ++ i)
           {
              if( comparand.get(i) > rv.get(i))
              {
                 rv.set(i, comparand.get(i));
              }
           }

           return rv;

        }

        Point<Dimensions> &operator+= (Point<Dimensions> const &comparand)
        {
           Point<Dimensions> &rv = *this;

           for(size_t i = 0; i < Dimensions; ++ i)
           {
               rv.dims_[i] += comparand.dims_[i];
           }

           return rv;

        }

        Point<Dimensions> &operator*= (double scaler)
        {
           Point<Dimensions> &rv = *this;

           for(size_t i = 0; i < Dimensions; ++ i)
           {
               rv.dims_[i] *= scaler;
           }

           return rv;

        }

        Point<Dimensions> &operator/= (double scaler)
        {
           Point<Dimensions> &rv = *this;

           for(size_t i = 0; i < Dimensions; ++ i)
           {
               rv.dims_[i] /= scaler;
           }

           return rv;

        }


    };  // class Point<Dimensions>

    //------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------

    template<class Ostream, int Dimensions>
    Ostream &operator<< (Ostream &s, Point<Dimensions> const &p)
    {
       p.output(s);

       return s;
    }

    //------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------

    template<class ListOfList, class Ostream>
    void logPointGroups(Ostream          &cout, 
                        ListOfList const &groups
                       )
         ///  print the points in a group to a stream
    {
        typedef typename ListOfList::value_type PointList;

        int group=0;

        CXXTLS_FOREACH( PointList const &cur, groups)
        {
           cout << group++ << '\n';

           cout << "    ";
        
           CXXTLS_FOREACH(typename PointList::value_type const &point, cur)
           {
              cout << point << " ";
           }
        
           cout << "\n\n";
        
        }
    }

} // namespace geometry
} // namespace cxxtls 
#endif
