#ifndef streamable_h_included
#define streamable_h_included
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file
///
///   This file defines the Ostreamable template and the StreableString class.
///
///   The Ostreamable<Derived> template's job is to
///   serve as a 'mixin' class.  That is, you derive from Ostreamable for the
///   purpose of having functions added into your class -- rather than for
///   for the purpose of defining your class to be a 'Ostreamable' object.
/// 
///   The class, StreamableString, adds 'Ostreamable' to 'std::string'
///   to give a quick and dirty way to construct formatted strings.  May not
///   be fast -- who knows.
/// 
/// 

#include <fstream>
#include <cxxtls/classTraits.h>

namespace cxxtls
{

template <class Owner, class charT=char, class traits = std::char_traits<charT> >
class OstreamableBuffer: public std::basic_streambuf<charT, traits>
    /// a trivial implementation of a stream buffer that only
    /// just routes the overflow() behavior (that prints characters)
    /// to class that owns the OstreamableBuffer
{
    Owner *owner;
public:
    typedef charT                          char_type;      
    typedef traits                         traits_type;
    typedef typename traits_type::int_type int_type;
    typedef typename traits_type::off_type off_type;
    typedef typename traits_type::pos_type pos_type;
        
        OstreamableBuffer(): owner(0) {}
        
        void setOwner(Owner *o) { owner = o; }

 protected:
    int_type overflow (int_type c = traits_type::eof ()); // calls Owner::ostreamable_helper() 
};

template <class Owner, class charT, class traits>
typename OstreamableBuffer<Owner, charT, traits>::int_type
OstreamableBuffer<Owner, charT, traits>::
overflow (typename OstreamableBuffer<Owner, charT, traits>::int_type c)
{
    // note that this is not a proper implemenation of overflow() for 
    // streams where you want to actually use the data in stream pointers.
    // This one should be copied with extreme caution.
    // See  http://stdcxx.apache.org/doc/stdlibug/39-2.html for a better
    // example, but still a wierd example...

    owner->ostreamable_helper(c);
    return c;
}

template <class Derived, class charT=char, class traits = std::char_traits<charT> >
class Ostreamable: public std::basic_iostream<charT, traits>
    ///
    /// A base class that hooks up a derived class to the ostream i/o operation
    /// set.  To give your class the ability to act like an ostream, derive from
    /// Ostreamble<Derived> and implement the ostreamable_helper(charT c) function
    /// that outputs a single character.  @see StreamableString
    ///
{
public:
    typedef charT                          char_type;    //2
    typedef traits                         traits_type;
    typedef typename traits_type::int_type int_type;
    typedef typename traits_type::off_type off_type;
    typedef typename traits_type::pos_type pos_type;

    typedef OstreamableBuffer<Derived, charT, traits> buf_type;
        
        Ostreamable();
        
        
private:

    buf_type buf;

public: 
        
};

template <class Derived, class charT, class traits>
Ostreamable<Derived, charT, traits>::
Ostreamable ()
: std::basic_iostream<charT, traits>(&buf)
{
    this->init (&buf); 

    Derived *owner=0;

    derivedFromBase(this, owner);

    buf.setOwner(owner);
}



///////////////////////////////////////////////////////////////////////////////////////

class StreamableString
  //
  /// This class implements a quick and dirty way to create formatted
  /// strings -- it is simpler to use that strstream but may be slower
  /// if you are working on large streams.  This kind of string is
  /// appropriate for quick formatting activities.
  //
: public Ostreamable<StreamableString>,
  public std::string
{
public:

    void ostreamable_helper(char c)
      ///  Helper that actually performs concatenations
    {
      std::string &p = *this;

      p += c;
    }

    StreamableString() 
         ///  Default constructor
    {
    } 

    StreamableString(StreamableString const &r)  ///  Copy constructor (only string content)
    {

      std::string &me = *this;  // SPOS MSVC 6.0
      me = r;
    }

    StreamableString( std::string const &r ) ///  Create from std::string
    {

      std::string &me = *this;
      me = r;
    }

    StreamableString( char const * r) ///  Create from C style string
    {
      std::string &me = *this;
      me = r;
    }

    StreamableString &operator=(std::string const &rhs)
    {
       std::string &me = *this;
       me = rhs;
       return *this;

    }

    StreamableString &operator=(char const *rhs)
    {
       std::string &me = *this;
       me = rhs;
       return *this;
    }


};

} // namespace cxxtls

#endif
