#ifndef format_h_header_included_p
#define format_h_header_included_p

///@file fmtio/fmtio.h
///
/// This file defines the class, fmtio::cs, that is used to help in formatting output or
/// strings using a printf style control string instead of using the iomanip objects.
/// It can also be used to populate std::strings with formatted text.
///
/// cs runs about 3 times faster than boost::format() but is not identical to it.

#include <portable_io.h>

#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>

// ========================================================================================

namespace fmtio
   /// A namespace encapsulating text formatting functions using control
   /// strings like those found in printf and doing it 3 times faster than
   /// boost's format class.
{

// ========================================================================================

namespace printableType
    /// namespace holding expression classifier types and methods. 
    ///
    /// There are two basic categories of symbols in this namespace
    ///
    ///  @li classifier() functions that determine what kind of expression is
    ///      passed to them.
    ///
    ///  @li data types whose sole purpose is to distinguish from another.
    ///
    /// The classifier functions return different return value types based on
    /// what kind of expression is passed to them.
    ///
{
  
// ========================================================================================

  //
  //  The following struct declarations are unique to the style of printing of the data
  //  associated with the struct names.
  //

  struct Integral      {};  ///< Identifies types printable as integers and pointers except char *s
  struct String        {};  ///< Identifies types printable as char const *, unsigned char const *
  struct StlString     {};  ///< Identifies types printable as string const &
  struct Float         {};  ///< Identifies types printable as float
  struct DoubleFloat   {};  ///< Identifies types printable as double
  struct UserDefined   {};  ///< Identifies types printable as all other types
  struct Unsigned      {};  ///< Identifies types printable as unsigned integral
  struct Char          {};  ///< Identifies types printable as single character


// ========================================================================================

  //
  // The printability of types does not depend on its constness, so we need a way
  // to eliminate const from type signatures.  Ditto for volatile.
  //



  template<class T> struct devolify             { typedef T type; };  ///< contains nested type, same as T but not volatile
  template<class T> struct devolify<T volatile> { typedef T type; };  ///< contains nested type, same as T but not volatile

  template<class T> struct deconstify           { typedef typename devolify<T>::type type; }; ///< contains nested type, same as T but not const or volatile
  template<class T> struct deconstify<T const>  { typedef typename devolify<T>::type type; }; ///< contains nested type, same as T but not const or volatile

// ========================================================================================

  template<int N> struct typeDetector;  
    ///<
    ///  This template struct defines a nested type, type, which is a typedef for one of the following:
    ///
    ///    UserDefinedType
    ///    Integral
    ///
    ///  The template is specialized for N=0 such that the nested type is Integral, and
    ///  for N=1 for UserDefined types (meaning classes)
    ///  
    ///

  template<> struct typeDetector<1> { typedef UserDefined type; }; // class objects
  template<> struct typeDetector<0> { typedef Integral    type; }; // enums

  template<class Type>
  struct isClassType
      ///
      /// Declares a const member, value, which is either 0 or 1 depending on 
      /// whether or not Type is a class object or not.
      ///
  {
    
    template <typename ClassType>
    static short fun(void (ClassType::*)());
    
    template <typename NonClassType>   // this has to be a template
    static char fun(...);
  
    enum { value = sizeof(fun<Type>(0)) == sizeof(short) };
  };



  template<class T>
  struct classification
      ///
      ///  Contains member, printableClass, that indicates the printing style of type T.
      ///
      ///  This instance of this class is meant ONLY to be used with class objects and enumeration
      ///  types.  All other types are specialized below so that this class can be simplier in
      ///  implementation than you might otherwise expect.
      ///  
      ///
  {
    typedef typename typeDetector< isClassType< typename deconstify<T>::type >::value >::type  printableClass;
  };



// ========================================================================================


  template<class T>
  typename classification< typename deconstify<T>::type >::printableClass 
  classifier(T const &t) 
    ///
    /// The template function, classifier<T>, returns a different type based on its parameter.
    ///
    /// It does this only because specializations for all builtin data types,
    /// change the default classification::printableClass data type, UserDefined,
    /// to one of the well known types (Integral, String, etc.).
    ///
    /// @param t specifies a reference to some expression value
    /// @return a data type that implies the printing algorithm for value
    ///
  { 
    return typename classification< typename deconstify<T>::type >::printableClass(); 
  }

// ========================================================================================

  template<class T> struct classification<T *>    { typedef Integral printableClass; }; 
     ///
     ///< Superseeds the default printing type category for pointers.
     ///
     /// Pointer types are conflated into integers for printing purposes.
     ///


// ========================================================================================

  template<> struct classification<int>           { typedef Integral printableClass; };  ///< integer class
  template<> struct classification<short>         { typedef Integral printableClass; };  ///< integer class
  template<> struct classification<long>          { typedef Integral printableClass; };  ///< integer class
  template<> struct classification<long long>     { typedef Integral printableClass; };  ///< integer class

  template<> struct classification<unsigned int>           { typedef Unsigned printableClass; };  ///< integer class
  template<> struct classification<unsigned short>         { typedef Unsigned printableClass; };  ///< integer class
  template<> struct classification<unsigned long>          { typedef Unsigned printableClass; };  ///< integer class
  template<> struct classification<unsigned long long>     { typedef Unsigned printableClass; };  ///< integer class


// ========================================================================================

  template<> struct classification<char>          { typedef Char     printableClass; };  ///< character class
  template<> struct classification<unsigned char> { typedef Char     printableClass; };  ///< character class

// ========================================================================================

  template<> struct classification<float>         { typedef Float       printableClass; }; ///< float class
  template<> struct classification<double>        { typedef DoubleFloat printableClass; }; ///< double class
  
// ========================================================================================

  // note:  g++ is weird about quoted strings -- they don't seem to match template name<char const *>
  // and they don't seem to match template name< char const (&)[N] >, so I have worked around this
  // in the cs class by making operator% handle char const * as a special case.  See below.

  template<> struct classification<char*>                { typedef String       printableClass; }; ///< string class
  template<> struct classification<unsigned char*>       { typedef String       printableClass; }; ///< string class
  template<> struct classification<unsigned char const *>{ typedef String       printableClass; }; ///< string class
  template<> struct classification<std::string>          { typedef StlString    printableClass; }; ///< string class

// ========================================================================================

  struct Exception
      ///<  Invalid format type specified for a given actual data type
  {
     char expectedFormat_;

     char const *foundType_;

     Exception(char e, char const *found)
     : expectedFormat_(e)
     , foundType_(found)
     {
     }

  };
};

// ========================================================================================

class cs
   ///
   /// The type of an expression that can be printed using operator <<.  
   ///
   /// This class behaves similarly to boost::format() but runs about 3 times faster.
   ///
   /// Expressions of this type are constructed like this:
   ///
   /// @code
   ///
   ///   fmtio::cs("%d-%d") % 13 % 14
   ///
   /// @endcode
   ///
   /// When printed using cout, you will see text of the form:
   ///
   /// @code
   ///
   ///   13-14
   ///
   /// @endcode
   ///
   /// You can also you expressions of this type to populate string variables:
   ///
   /// @code
   ///
   ///  string fred = fmtio::cs("%-04x") % 8;
   ///
   /// @endcode
   ///
   /// In this case, fred is initialized with 0008.
   ///
   /// Variables of this type are only minimally useful -- the general idea is that
   /// you use the constructor, cs(char const *), as passed in an cout expression
   /// to construct formatted output in a manner that is easier to use than if you had used 
   /// the iomanip header file contents to do the same job. 
   ///
   /// For example:
   ///
   ///   cout << fmtio::cs("Stuff:  %-08.4f") % 9.3 << endl;
   ///
   /// In this case, the sub-expression, cs(...) % 9.3, is converted into a string
   /// which is then printed as:  Stuff:  0009.300.
   ///
   /// Most builtin types can all be sent to ostreams or converted into strings by this class.
   ///
   /// User defined struct and classes can only be processed if there is an operator<<(std::ostream,...)
   /// defined for them.
   ///
   /// Pointers and arrays are automatically treated as integers.  You can use %x to print them as
   /// hex.
   ///
   /// Enumerations are not at this time usable.  Convert them to integers.  For example:
   ///
   /// @code
   ///   #include <iostream>
   ///   #include <fmtio/fmtio.h>
   ///
   ///   enum stuff { enum1 };
   ///
   ///   int main()
   ///   {
   ///      std::cout << fmtio::cs("%d") % enum1 << std::endl;  // WILL NOT COMPILE!
   ///
   ///      std::cout << fmtio::cs("%d") % int(enum1) << std::endl;  // works ok
   ///      
   ///   }
   /// @endcode
   ///
   /// The control strings are simplified form of printf -- basically all you get is this:
   ///
   ///  @li - if used must be first to specify left justification
   ///  @li 0 if used must be before any field width to specify that 0 is the fill character
   ///  @li fieldWidth if used, specifies the maximum width of the field
   ///  @li . if used must appear before any precision 
   ///  @li precision  if used, specifies the width of the remaining part of the text to be
   ///      converted
   ///  @li converter required character that indicates what form of conversion is going on
   ///
   ///  sprintf is used to generate most of the strings, but floats and and user defined types
   ///  are printed using operator<< into a stringstream.
   ///
   /// The conversion characters are as follows:
   ///
   ///   @li d produces signed decimal output
   ///   @li u produces unsigned decimal output
   ///   @li x produces hexadecimal (unsigned) output
   ///   @li o produces octal output
   ///   @li f produces floating point output (float and double and long double)
   ///   @li U produces user defined type output (requires the user to have an operator<< for the type)
   /// 
   /// Multiple % operators can be used with a given cs() but multiple % conversions
   /// must be specified when this occurs and they must correcly the match the basic data
   /// type as detected by the classifer() methods.  Here's an example of one that works:
   ///
   /// @code
   ///    cout << fmtio::cs("%s %d %x")  %"stringdata" %integerVariable %integerVariable << endl;
   /// @endcode
   ///
   /// And here is an invalid example:
   ///
   /// @code
   ///    cout << fmtio::cs("%s %d %x")  %1.0 %integerVariable %integerVariable << endl;
   /// @endcode
   ///
   /// In this invalid example, %s is specified but a floating point data type is actually
   /// supplied for the first parameter.
   /// 
   /// All the parameters must be matched and they must all be correct.
   ///
{
    // ---------------------------------------------------------------------------------------

     mutable char expectedFormat_;  // %<expectedFormat_> in control string language

     mutable char const *cs_;

     mutable std::string buffer_;

     mutable int fieldWidth_;
     mutable int fillChar_;
     mutable int precision_;
     mutable int left_;

    // ---------------------------------------------------------------------------------------

     void reset() const
         ///  Reset the field formatting fields to their default state.
     {
       left_       = 0;
       fieldWidth_ = 0;
       fillChar_   = 0;
       precision_  = 0;
     }


public:

    // ---------------------------------------------------------------------------------------

     operator std::string const &() const { return buffer_; }

    // ---------------------------------------------------------------------------------------

     cs(char const *cs)
         ///
         ///  Begin a conversion sequence that contains one or more % operators to define the
         ///  variables of interest.  For example:
         ///
         ///  @code
         ///    cout << cs("%d, %d") %100 %200 << endl;
         ///  @endcode
         ///
         ///  Will produce "100, 200" to sdout
         ///
         ///
     : expectedFormat_(*cs)
     , cs_(cs)
     {
       advanceFormat(); 
     }

     ~cs()
     {
       if(expectedFormat_ != '?' )
       {
         // should have been out of formatting directives by now

         throw  printableType::Exception(expectedFormat_, "<EXPECTED END OF FORMAT STRING>");  

       }
     }

private:

    // ---------------------------------------------------------------------------------------

     void advanceFormat() const
         ///
         /// Consume text from the control string until the next '%' character and append
         /// the text to the output buffer.
         ///
         /// This function treats %% as a %, per printf standards.
         ///
     {
       reset();  // field formatting options

       // skip non-format text before % and handle %%

       while(*cs_)                     // outer loop to handle repeats after %%
       {

         while(*cs_ && *cs_ != '%')   // normal case -- just print anything not a % spec
           buffer_ += *cs_++;
  
         if(*cs_ == '%')              // ignore the % of a spec
           ++cs_;

         if(*cs_ == '%')
         {
           buffer_ += '%';
           ++cs_;
         }
         else
           break;
       }

       // the leading % is already removed before we get here

       expectedFormat_ = '?';  // set up an invalid format as the norm -- superseed
                               // the invalid state only if a properly formatted
                               // control string is found.

       if(*cs_ == '-')
       {
         left_ = 1;            // handle left justification
         ++cs_;
       }

       if(*cs_ == '0')
       {
         fillChar_ = '0';      // handle leading zeros
         ++cs_;
       }

       if(*cs_ >= '0' && *cs_ <='9')
       {
         fieldWidth_ = 0;                    // handle field widths 

         while(*cs_ >= '0' && *cs_ <= '9')
         {
           fieldWidth_ *= 10;
           fieldWidth_ += *cs_++ -'0';
         }
       }

       if(*cs_ == '.')
       {
         ++cs_;                             // handle precision

         precision_ = 0;

         while(*cs_ >= '0' && *cs_ <= '9')
         {
           precision_ *= 10;
           precision_ += *cs_++ -'0';
         }

       }

       // now, handle the formatting character:  u, x, d, o, s, c, U (U is user defined)

       if(*cs_)
       {
         expectedFormat_ = *cs_++;
       }
       
     }


    // ---------------------------------------------------------------------------------------
     void outputIntegral( long long t ) const
         ///
         ///  General purpose integer output routine
         ///
     {
         char b[100];

         char cs[40];

         char *scan = &cs[0];

         {
  
           *scan++ = '%';  // setup control string for converison
  
           if(left_)
             *scan++ = '-';

           if(fillChar_ == '0')
             *scan++ = '0';

           if(fieldWidth_)
           {
             char buffer[40];

             snprintf(buffer, 40, "%d", fieldWidth_);

             char *input = &buffer[0];

             while(*input)
               *scan++ = *input++;

           }

           if(precision_)
           {
             *scan++ = '.';

             char buffer[40];

             snprintf(buffer, 40, "%d", precision_);

             char *input = &buffer[0];

             while(*input)
               *scan++ = *input++;

           }
         }

         *scan++ = 'l';                  // always use long long output data (see cast below)
         *scan++ = 'l';
         *scan++ = expectedFormat_;
         *scan++ = 0;


         snprintf(b, sizeof(b), cs, (long long)t);

         buffer_ += b;

     }

    // ---------------------------------------------------------------------------------------
     bool isIntegralFormat() const
         ///
         ///  Return true if the current expectedFormat_ is an integral format.
         ///
     {
       return    expectedFormat_ == 'd'
              || expectedFormat_ == 'u'
              || expectedFormat_ == 'o'
              || expectedFormat_ == 'x'
              ;
     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::Integral) const
         ///
         ///  Used to output signed integral data types like int, short, long, unsigned, long long, etc
         ///

     { 
         if( !isIntegralFormat() )
         {
           throw  printableType::Exception(expectedFormat_, "Integral");  
         }

         outputIntegral((long long)t);

     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::Unsigned)    const 
         ///
         ///  Used to output unsigned integral data types like unsigned int, unsigned short, unsigned char, etc.
         ///
     { 
         if( !isIntegralFormat() )
         {
           throw  printableType::Exception(expectedFormat_, "Unsigned");  
         }

         outputIntegral((long long)t);

     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::Char)        const 
         ///
         ///  Used to print character data.
         ///
     {
         if(expectedFormat_ == 'c')
         {
              buffer_ += (t);
         }
         else
         if( !isIntegralFormat() )
         {
           throw  printableType::Exception(expectedFormat_, "Char");
         }
         else
           outputIntegral(t);
     }


    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::String)      const 
        ///
        ///  Used to print char const *, and unsigned char const *.
        ///
     { 
         if(   expectedFormat_ != 's'
           )
         {
           throw  printableType::Exception(expectedFormat_, "String");
         }

       
         buffer_ += (char const *)(t);
     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::StlString)   const 
        ///
        ///  Used to print std::string
        ///
     { 
         if(   expectedFormat_ != 's'
           )
         {
           throw  printableType::Exception(expectedFormat_, "std::string");
         }


         buffer_ += t;

     }

    // ---------------------------------------------------------------------------------------
     void outputFloat(double d) const
         ///
         ///  Generalized floating point print routine
         ///
     {
         std::stringstream tmp;

         if(left_)
         {
           tmp << std::left;
         }

         if(fillChar_)
         {
           tmp << std::setfill(char(fillChar_));
         }

         if(fieldWidth_)
         {
             tmp << std::setw(fieldWidth_);
         }

         if(precision_)
         {
             tmp << std::setprecision(precision_);
         }
         
         tmp << d;

         buffer_ +=tmp.str();
     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::Float)       const 
         ///
         ///  Used to print floats
         ///
     { 
         if(   expectedFormat_ != 'f'
           )
         {
           throw  printableType::Exception(expectedFormat_, "Float");
         }

         outputFloat(t);

     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::DoubleFloat) const 
        ///
        ///  Used to print doubles
        ///
     { 
         if(   expectedFormat_ != 'f'
           )
         {
           throw  printableType::Exception(expectedFormat_, "DoubleFloat");
         }

         outputFloat(t);

     }

    // ---------------------------------------------------------------------------------------
     template<class T>  void output(T const &t, printableType::UserDefined) const 
         ///
         ///  Used to print user defined types
         ///
     { 
         if(   expectedFormat_ != 'U'
           )
         {
           throw  printableType::Exception(expectedFormat_, "UserDefined");
         }

         std::stringstream tmp;

         if(left_)
         {
           tmp << std::left;
         }

         if(fillChar_)
         {
           tmp << std::setfill(char(fillChar_));
         }

         tmp << std::setw(fieldWidth_) << std::setprecision(precision_) << t;

         buffer_ += tmp.str();
     }

public:

    // ---------------------------------------------------------------------------------------
     template< class T >
     cs const &operator% ( T const &rhs ) const
         /// Used to print any type.  Triggered by the following syntax is used:
         ///
         /// @code
         ///
         ///  cout << cs(cs) % someObject;
         ///
         /// @endcode
         ///
         /// This member assumes that the cs will have been constructed with the
         /// appropriate % conversion character in the correct position for this "%someObject"
         /// to match it.
         ///

     {
        output(rhs, printableType::classifier(rhs));

        advanceFormat();

        return *this;
     }

    // ---------------------------------------------------------------------------------------
     cs const &operator% ( char const *rhs ) const
         /// Used to print any type.  Triggered by the following syntax is used:
         ///
         /// @code
         ///
         ///  cout << cs(cs) %"[";
         ///
         /// @endcode
         ///
         /// This member assumes that the cs will have been constructed with the
         /// appropriate % conversion character in the correct position for this "%someObject"
         /// to match it.
         ///

     {
        output(rhs, printableType::String() );

        advanceFormat();

        return *this;
     }

    // ---------------------------------------------------------------------------------------
     template<class T, int N>
     cs const &operator% ( T const (&rhs)[N] ) const
         /// Used to print any type.  Triggered by the following syntax is used:
         ///
         /// @code
         ///
         ///  long a_long[40];
         ///
         ///  cout << cs(cs) %a_long;
         ///
         /// @endcode
         ///
         /// This member assumes that the cs will have been constructed with the
         /// appropriate % conversion character in the correct position for this "%someObject"
         /// to match it.
         ///

     {
        output(rhs, printableType::Integral() );

        advanceFormat();

        return *this;
     }


    size_t size() const { return buffer_.size(); }

private:

    // ---------------------------------------------------------------------------------------

    cs(cs const &);  // not implemented:  If you get an error on this line of code, 
                     // you probably wrote code that looks like this:  
                     //
                     //     cout << cs("...") << something << endl;
                     //                             ^^^^
                     //  No can do big guy!  Use %, not <<



     template< class T > void operator<< ( T & ) const;   // not implemented

};

// ===========================================================================================

inline std::ostream &operator<< (std::ostream &s, cs const &fmt)
{
   if(s.good())
   {
   
      s << (std::string const &) fmt;

   }

   return s;
}

// ===========================================================================================
// ===========================================================================================

}
#endif
