#ifndef script_command_header_included_p
#define script_command_header_included_p
//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/scriptthread.h>
#include <cxxtls/threadPool.h>
#include <cxxtls/scriptfunctions.h>
#include <utility>

///@file
///   This file defines a multi-threaded command handler -- at least the core components:
///   @li a definition of what a command is.
///   @li a pool of threads that executes commands of that type
///   @li a scheme for knowing when to end execution normally.

namespace cxxtls
{

template<class T>
struct GuardedCounter
    ///  An integer counter that can be incremented and decremented.  It can also checked for specific values
    /// after the 
{
    T value_;

    Mutex guard_;

    GuardedCounter(T const &t)
    {
        ScopedMutexLock lock(guard_);
        value_ = t;
    }

    void set(T const &t)
    {
        ScopedMutexLock lock(guard_);
        value_ = t;
    }

    operator T() 
    {
        ScopedMutexLock lock(guard_);

        return value_;
    }

    void inc()
    {
        ScopedMutexLock lock(guard_);
        ++value_;
    }

    void dec()
    {
        ScopedMutexLock lock(guard_);
        --value_;
    }

    bool decAndTest(T const &value)
    {
        ScopedMutexLock lock(guard_);
        --value_;

        return value_ == value;
    }

    bool incAndTest(T const &value)
    {
        ScopedMutexLock lock(guard_);
        ++value_;

        return value_ == value;
    }


};


struct ScriptCommandHandler;  // predefine

class ScriptCommand
    // a value object that encapsulates the entirety of the commands that can be sent
    // to a thread pool.
    //
    // To send a command, new up a ScriptCommand Object, then push it into the global thread
    // pool for execution.  The pool will eventually invoke the execute() method which
    // will call a derived class run() and when it finishes, the the execute method will
    // delete self...
    // 
    // To implement a script command, derive from this class and superseed the run() function
    //
    // NOTE:  A ScriptCommand is not a statement in a script language.  A script command 
    // would be the entirety of information needed to execute a script:  environment, compiled
    // functions, file handles, the whole enchilada.  Multiple script language statements are
    // likely to be executed in a single script command.  Each possibly with its on ScriptCommand
    // object.  
    //
    // ScriptCommands are more like "job ids" than they are statements.
{
protected:
    ScriptCommandHandler *owner_;

public:

    inline ScriptCommand(ScriptCommandHandler *owner); 


    inline virtual ~ScriptCommand();

    virtual void run() = 0;

    inline void shutdown();

};


struct ScriptCommandHandler
:  public cxxtls::ThreadCommandHandler<ScriptCommand*>
///   A combination of a thread pool that executes ScriptCommand* objects and also
///   the algorithm for implementing the command handling itself.
///
///   You get two birds with one stone with this baby.
{

   typedef cxxtls::ThreadCommandHandlerPool<ScriptCommand*>  ScriptCommandPool;

   std::auto_ptr<ScriptCommandPool> threadPool_;

   GuardedCounter<long> activeCommands_;

   cxxtls::ThreadQueue<int> shutdownRequested_;

   friend class ScriptCommand;

   ScriptCommandHandler(int maxThreads)
   :  activeCommands_(0)
   {
       threadPool_.reset(new ScriptCommandPool(this, maxThreads));

   }

    
    virtual bool handle(ScriptCommand* &cmd) const
    {
       cmd->run();

       delete cmd;

       return false;
        
    }

    void push(ScriptCommand* cmd)  // push a command into input queue
    {
        threadPool_->push(cmd);
    }


    void waitForAllThreadsToTerminate()
    {
        int unusedVariable=0;
        shutdownRequested_.waitPop(unusedVariable);
        threadPool_->shutdown();   // when last ScriptCommand is destroyed, this will terminate
    }

    bool terminated() 
    {
       return threadPool_->terminated();
    }
    
};

inline
ScriptCommand::
ScriptCommand(ScriptCommandHandler *owner)
:  owner_(owner)
{
    owner_->activeCommands_.inc();
}


inline 
ScriptCommand::
~ScriptCommand()
{
   bool outOfCommands = owner_->activeCommands_.decAndTest(0); 

   if(outOfCommands)
   {
       owner_->shutdownRequested_.push(0);  // any word will do
   }
}

inline
void
ScriptCommand::
shutdown()
{
    owner_->shutdownRequested_.push(0);  // any word will do
}

}  // namspace cxxtls
#endif
